/* !!! To test:
  Haan: square is destroyed (FS) when a piece leaves it (no castling)
  Null: square is destroyed (FS) when a capture occurs on it
  (anti)Hypervolages: a piece changes color when changing of square color
  Frankfurter, Pocket Knight, Kamikaze, Magnetic, Einstein, Gravitation,
  Banana skin, Extinction ... */

/* Search for (short) uniquely realisable mate in chess variants */
/* Alain Brobecker */
/* 2008.08.13 - added function for printing moves
                added mate/stalemate recognition */
/* 2008.08.12 - added move generation for kNight, with a macro */
/* 2008.05.23 - modified IsKingInCheck(i) to IsSquareAttackedBy(i,c) so that
                it can be used to check castling */
/* 2008.05.18 - added IsKingInCheck(i) */
/* 2008.05.17 - start+speed tests: dim1 arrays are faster than dim2 */
#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

#define MaxPly 2
/* c is for color (0=White, 1=Black), must be lowest bit. */
/* 2nd bit is set for pieces. And Q=B+R. */
#define ES 0  /* 000000 empty square */
#define FS 1  /* 000001 forbidden squares around the board */
#define WK 2  /* 00001c K */
#define BK 3
#define WP 6  /* 00011c P */
#define BP 7
#define WN 10 /* 00101c N */
#define BN 11
#define WB 18 /* 01001c B */
#define BB 19
#define WR 34 /* 10001c R */
#define BR 35
#define WQ 50 /* 11001c Q */
#define BQ 51




/* Every Board is made of 21+8+2+8+2+8+2+8+2+8+2+8+2+8+2+8+21=120 chars */
/* The 21s and 2s are used for forbidden squares (FS) around the board */
unsigned char Board[(MaxPly+1)*12*10];
unsigned char ep[MaxPly+1]; /* square where and ep capture is possible or 0 */
unsigned char castle[MaxPly+1]; /* 1 bit for every K or R */
long int NbGames[MaxPly+1];

/* macro used to generate moves for K,N */
#define GenerateMove(Displacement,OppositeColorTest,t) \
  t=Board[i+Displacement]; \
  if((t==0) || ((t & (1+2))==OppositeColorTest)) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+Displacement+120]=Board[i]; \
    Board[i+120]=0; \
    switch(RecursiveSearch(Ply+1)) { \
      case -1: break; \
      case 1: PrintMove(i,Displacement,'#'); NbLegalMoves++; break; \
      case 2: PrintMove(i,Displacement,'='); PrintPly(Ply+1); NbLegalMoves++; break; \
      case 0: PrintMove(i,Displacement,' '); NbLegalMoves++; break; \
    } \
  }

/* Prints move from square index to square index+displacement */  
void PrintMove(Index,Displacement,EndChar) int Index,Displacement; char EndChar; {
  int Ply,temp;
  Ply=Index/120;
  printf("%d.",(Ply/2)+1);
  if((Ply&1)==1) { printf(".. "); }
  switch(Board[Index]&254) {
    case WK: printf("K"); break;
    case WP: printf("P"); break;
    case WN: printf("N"); break;
    case WB: printf("B"); break;
    case WR: printf("R"); break;
    case WQ: printf("Q"); break;
    default: printf("***error1***");
  }
  temp=Index-120*Ply-21;
  printf("%c%d",temp%10+'a',8-temp/10);
  if(Board[Index+Displacement]==0) {
    printf("-");
  } else {
    printf("x");
    switch(Board[Index+Displacement]&254) {
      case WK: printf("K"); break;
      case WP: printf("P"); break;
      case WN: printf("N"); break;
      case WB: printf("B"); break;
      case WR: printf("R"); break;
      case WQ: printf("Q"); break;
      default: printf("***error2***");
    }
  }
  temp=Index+Displacement-120*Ply-21;
  printf("%c%d",temp%10+'a',8-temp/10);
  printf("%c\n",EndChar);
}
    

/* Print 8*8 central part of Board[Ply] */
void PrintPly(Ply) int Ply; {
  int i,x,y;
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i++]) {
        case ES: printf("."); break; case FS: printf("*"); break;
        case WK: printf("K"); break; case BK: printf("k"); break;
        case WP: printf("P"); break; case BP: printf("p"); break;
        case WN: printf("N"); break; case BN: printf("n"); break;
        case WB: printf("B"); break; case BB: printf("b"); break;
        case WR: printf("R"); break; case BR: printf("r"); break;
        case WQ: printf("Q"); break; case BQ: printf("q"); break;
        default:printf("?"); break; /* error */
      }
    }
    printf("\n");
    i+=2;
  }
}

/* Copy Board[PlyA] to Board[PlyB] */
void CopyPlyAToPlyB(PlyA,PlyB) int PlyA,PlyB; {
  int i,j,x,y;
  j=120*PlyB+21; i=120*PlyA+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) { Board[j++]=Board[i++]; }
    j+=2; i+=2;
  }
  castle[PlyB]=castle[PlyA]; /* copy castling rights (but not ep) */
  /*!!! set ep to 0??? */
  ep[PlyB]=0;
}

/**************************************************************************/
/* Returns -1 if square "index" is attacked by a "color" piece, else 0.   */
int IsSquareAttackedBy(index,color) int index; {
  int i;
  /* attacked by opposite color N? */
  if(Board[index-8]==(WN | color)) { return(-1); }
  if(Board[index-12]==(WN | color)) { return(-1); }
  if(Board[index-19]==(WN | color)) { return(-1); }
  if(Board[index-21]==(WN | color)) { return(-1); }
  if(Board[index+21]==(WN | color)) { return(-1); }
  if(Board[index+19]==(WN | color)) { return(-1); }
  if(Board[index+12]==(WN | color)) { return(-1); }
  if(Board[index+8]==(WN | color)) { return(-1); }
  /* attacked by opposite color P? */
  if(Board[index+9-18*color]==(WP | color)) { return(-1); }
  if(Board[index+11-22*color]==(WP | color)) { return(-1); }
  /* attacked by opposite color B or Q? */
  i=index+11; while(Board[i]==ES) { i+=11; }
  if((Board[i] & BB)==(WB | color)) { return(-1); }
  i=index+9; while(Board[i]==ES) { i+=9; }
  if((Board[i] & BB)==(WB | color)) { return(-1); }
  i=index-9; while(Board[i]==ES) { i-=9; }
  if((Board[i] & BB)==(WB | color)) { return(-1); }
  i=index-11; while(Board[i]==ES) { i-=11; }
  if((Board[i] & BB)==(WB | color)) { return(-1); }
  /* attacked by opposite color R or Q? */
  i=index+10; while(Board[i]==ES) { i+=10; }
  if((Board[i] & BR)==(WR | color)) { return(-1); }
  i=index+1; while(Board[i]==ES) { i+=1; }
  if((Board[i] & BR)==(WR | color)) { return(-1); }
  i=index-1; while(Board[i]==ES) { i-=1; }
  if((Board[i] & BR)==(WR | color)) { return(-1); }
  i=index-10; while(Board[i]==ES) { i-=10; }
  if((Board[i] & BR)==(WR | color)) { return(-1); }
  return(0);
}

/**************************************************************************/
/* Returns:
  -1 if K of previous player is still in check (previous move was illegal)
   1 if K of current player is mate
   2 if K of current player is stalemate 
   0 otherwise. */
int RecursiveSearch(Ply) int Ply; {
  int i,x,y,WKx,WKy,WKi,BKx,BKy,BKi,NbLegalMoves,temp,OppositePlayer;
  /* check that distance between both kings is >1 */
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i]) {
        case WK: WKx=x; WKy=y; WKi=i; break;
        case BK: BKx=x; BKy=y; BKi=i; break;
      }
      i++;
    }
    i+=2;
  }
  x=WKx-BKx; if(x<0) { x=-x; }
  y=WKy-BKy; if(y<0) { y=-y; }
  if((x<2) && (y<2)) { return(-1); } /*!!!printf("dist WK-BK: (%d;%d)\n",x,y);*/
  
  /* make sure that K of previous player is not in check */
  if((Ply & 1)==0) { i=BKi; } else { i=WKi; }
  if(IsSquareAttackedBy(i,Ply & 1)==-1) { return(-1); }
  
  /* if we are not at MaxPly, generate all possible moves */
  if(Ply<MaxPly) {
    OppositePlayer=1-(Ply&1)+2;
    NbLegalMoves=0;
    i=120*Ply+21;
    for(y=0;y<8;y++) {
      for(x=0;x<8;x++) {
        /* Board[i] contains a piece of current's player color? */
        if((Board[i]&(2+1))==(2+(Ply&1))) {
          switch(Board[i]&254) {
            /* generate moves for K */
            case WK:
              GenerateMove(-1,OppositePlayer,temp)
              GenerateMove(-11,OppositePlayer,temp)
              GenerateMove(-10,OppositePlayer,temp)
              GenerateMove(-9,OppositePlayer,temp)
              GenerateMove(+1,OppositePlayer,temp)
              GenerateMove(+11,OppositePlayer,temp)
              GenerateMove(+10,OppositePlayer,temp)
              GenerateMove(+9,OppositePlayer,temp)
              break;
            /* generate moves for P */
            /*!!! DOUBLE STEP + PROMOTION + EP + ACTUALISE EP RIGHTS */
            case WP: /*printf("P");*/ break;
            /* generate moves for N */
            case WN:
              GenerateMove(-8,OppositePlayer,temp)
              GenerateMove(-12,OppositePlayer,temp)
              GenerateMove(-19,OppositePlayer,temp)
              GenerateMove(-21,OppositePlayer,temp)
              GenerateMove(+21,OppositePlayer,temp)
              GenerateMove(+19,OppositePlayer,temp)
              GenerateMove(+12,OppositePlayer,temp)
              GenerateMove(+8,OppositePlayer,temp)
              break;
            default: 
              /* generate moves for B */
              if((Board[i]&WB)==WB) { /*printf("B");*/
              }
              /* generate moves for R */
              if((Board[i]&WR)==WR) { /*printf("R");*/
              }
            break;
          }
        }
        i++;
      }
      i+=2;
    }
    /*!!! castlings, will use IsSquareAttackedBy(i,c) */
    /* !!! printf("nb of legal moves:%d\n",NbLegalMoves);  */

    /* check if previous move mated, stalemated or nothing */  
    if(NbLegalMoves==0) {
      /* make sure that K of current player is not in check */
      if((Ply & 1)==0) { i=WKi; } else { i=BKi; }
      if(IsSquareAttackedBy(i,1-(Ply & 1))==-1) {
        return(1);
      } else {
        return(2);
      }
    } else {
      return(0);
    }
  }
  return(0);
}

/**************************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  int c,i,x,y;
  setbuf(stdout, NULL); /* no buffering for output! */
  /* initialise arrays */
  for(c=0;c<MaxPly+1;c++) {
    ep[c]=0;      /* no ep capture possible */
    castle[c]=63; /* no K or R has moved */
    NbGames[c]=0;
  }
  /* initialise Board for Ply 0 */
  for(c=0;c<(MaxPly+1)*12*10;c++) { Board[c]=FS; } /* forbidden */
  Board[21]=BR;Board[22]=BN;Board[23]=BB;Board[24]=BQ;
  Board[25]=BK;Board[26]=BB;Board[27]=BN;Board[28]=BR;
  for(x=0;x<8;x++) { Board[31+x]=ES; /*BP;*/ }  
  i=41;for(y=2;y<6;y++){for(x=0;x<8;x++){Board[i++]=ES;}i+=2;} /* empty */
  for(x=0;x<8;x++) { Board[81+x]=ES; /*WP;*/ }  
  Board[91]=WR;Board[92]=WN;Board[93]=WB;Board[94]=WQ;
  Board[95]=WK;Board[96]=WB;Board[97]=WN;Board[98]=WR;
  
/*!!!  Board[62]=BK; Board[73]=BQ;*/
  
  PrintPly(0);
  printf("\n");
 
  RecursiveSearch(0);
  
  exit(0);
}
