/* Search for (short) uniquely realisable mate in chess variants */
/* Alain Brobecker - 2008 august */
/* 2008.08.17 - set WKx,WKy,BKx,BKy as global variables */
/* 2008.08.15 - added (simple) recognition of similar EndMoves */
/* 2008.08.14 - added move generation for Pawns (ep but no promotion) */
/* 2008.08.13 - added function for printing moves
                added mate/stalemate recognition
                added move generation for Kings, Rooks, Bishops, Queens */
/* 2008.08.12 - added move generation for kNights, with a macro */
/* 2008.05.23 - modified IsKingInCheck(i) to IsSquareAttackedBy(i,c) so that
                it can be used to check castling */
/* 2008.05.18 - added IsKingInCheck(i) */
/* 2008.05.17 - start+speed tests: dim1 arrays are faster than dim2 */
#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

#define MaxPly 6
/* c is for color (0=White, 1=Black), must be lowest bit. */
/* 2nd bit is set for pieces. And Q=B+R. */
#define ES 0  /* 000000 empty square */
#define FS 1  /* 000001 forbidden squares around the board */
#define WK 2  /* 00001c K */
#define BK 3
#define WP 6  /* 00011c P */
#define BP 7
#define WN 10 /* 00101c N */
#define BN 11
#define WB 18 /* 01001c B */
#define BB 19
#define WR 34 /* 10001c R */
#define BR 35
#define WQ 50 /* 11001c Q */
#define BQ 51


/* Every Board is made of 21+8+2+8+2+8+2+8+2+8+2+8+2+8+2+8+21=120 chars */
/* The 21s and 2s are used for forbidden squares (FS) around the board */
unsigned char Board[(MaxPly+1)*12*10];
unsigned char ep[MaxPly+2]; /* square where an ep capture is possible or 0 */
unsigned char castle[MaxPly+1]; /* 1 bit for every K or R */
double NbGames[MaxPly+1];
int NbMates[MaxPly+1],NbStalemates[MaxPly+1];
unsigned int EndMoves[2*100000],NbEndMoves=0;


/* Insert move from square index to square index+displacement into list */
unsigned int IM_Ply,IM_Move,IM_t,IM_Found;
void InsertMove(Index,Displacement,Type) int Index,Displacement,Type; {
  IM_Ply=Index/120;
  IM_Move=IM_Ply<<26;                      /* ply (6 bits) */
  IM_Move+=(Board[Index]&254)<<(21-1);  /* piece (5 bits) */
  IM_Move+=(Index-120*IM_Ply)<<14;      /* square in [0;119] (7 bits) */
  IM_Move+=(Board[Index+Displacement]&254)<<(9-1); /* piece (5 bits) */
  IM_Move+=(Index+Displacement-120*IM_Ply)<<2; /* square (7 bits) */
  IM_Move+=Type; /* 0=mate, 1=stalemate, 2=ep mate, 3=ep stalemate (2 bits) */
  IM_Found=0;
  IM_t=0;
  while((IM_t<NbEndMoves) && (IM_Found==0)) {
    if(IM_Move==EndMoves[2*IM_t]) {
      IM_Found=-1;
      EndMoves[2*IM_t+1]++;
    }
    IM_t++;
  }
  if(IM_Found==0) {
    EndMoves[2*NbEndMoves]=IM_Move;
    EndMoves[2*NbEndMoves+1]=1;
    NbEndMoves++;    
  }
}

/* Print 8*8 central part of Board[Ply] */
void PrintPly(Ply) int Ply; {
  int i,x,y;
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i++]) {
        case ES: printf("."); break; case FS: printf("*"); break;
        case WK: printf("K"); break; case BK: printf("k"); break;
        case WP: printf("P"); break; case BP: printf("p"); break;
        case WN: printf("N"); break; case BN: printf("n"); break;
        case WB: printf("B"); break; case BB: printf("b"); break;
        case WR: printf("R"); break; case BR: printf("r"); break;
        case WQ: printf("Q"); break; case BQ: printf("q"); break;
        default:printf("?"); break; /* error */
      }
    }
    printf("\n");
    i+=2;
  }
}

/* Copy Board[PlyA] to Board[PlyB] */
void CopyPlyAToPlyB(PlyA,PlyB) int PlyA,PlyB; {
  int i,j,x,y;
  j=120*PlyB+21; i=120*PlyA+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) { Board[j++]=Board[i++]; }
    j+=2; i+=2;
  }
  castle[PlyB]=castle[PlyA]; /* copy castling rights */
}

/**************************************************************************/
/* Returns -1 if square "index" is attacked by a "color" piece, else 0.   */
int IsSquareAttackedBy(index,color) int index; {
  int i;
  /* attacked by opposite color N? */
  if(Board[index-8]==(WN | color)) { return(-1); }
  if(Board[index-12]==(WN | color)) { return(-1); }
  if(Board[index-19]==(WN | color)) { return(-1); }
  if(Board[index-21]==(WN | color)) { return(-1); }
  if(Board[index+21]==(WN | color)) { return(-1); }
  if(Board[index+19]==(WN | color)) { return(-1); }
  if(Board[index+12]==(WN | color)) { return(-1); }
  if(Board[index+8]==(WN | color)) { return(-1); }
  /* attacked by opposite color P? */
  if(Board[index+9-18*color]==(WP | color)) { return(-1); }
  if(Board[index+11-22*color]==(WP | color)) { return(-1); }
  /* attacked by opposite color B or Q? */
  i=index+11; while(Board[i]==ES) { i+=11; }
  if((Board[i] & BB)==(WB | color)) { return(-1); }
  i=index+9; while(Board[i]==ES) { i+=9; }
  if((Board[i] & BB)==(WB | color)) { return(-1); }
  i=index-9; while(Board[i]==ES) { i-=9; }
  if((Board[i] & BB)==(WB | color)) { return(-1); }
  i=index-11; while(Board[i]==ES) { i-=11; }
  if((Board[i] & BB)==(WB | color)) { return(-1); }
  /* attacked by opposite color R or Q? */
  i=index+10; while(Board[i]==ES) { i+=10; }
  if((Board[i] & BR)==(WR | color)) { return(-1); }
  i=index+1; while(Board[i]==ES) { i+=1; }
  if((Board[i] & BR)==(WR | color)) { return(-1); }
  i=index-1; while(Board[i]==ES) { i-=1; }
  if((Board[i] & BR)==(WR | color)) { return(-1); }
  i=index-10; while(Board[i]==ES) { i-=10; }
  if((Board[i] & BR)==(WR | color)) { return(-1); }
  return(0);
}

/**************************************************************************/
/* Returns:
  -1 if K of previous player is still in check (previous move was illegal)
   1 if K of current player is mate
   2 if K of current player is stalemate 
   0 otherwise. */
int WKx,WKy,BKx,BKy;
int RecursiveSearch(Ply) int Ply; {
  int i,x,y,WKi,BKi,NbLegalMoves,temp,OppositePlayer,d;
  /* check that distance between both kings is >1 */
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i]) {
        case WK: WKx=x; WKy=y; WKi=i; break;
        case BK: BKx=x; BKy=y; BKi=i; break;
      }
      i++;
    }
    i+=2;
  }
  x=WKx-BKx; if(x<0) { x=-x; }
  y=WKy-BKy; if(y<0) { y=-y; }
  if((x<2) && (y<2)) { return(-1); } /*!!!printf("dist WK-BK: (%d;%d)\n",x,y);*/
  
  /* make sure that K of previous player is not in check */
  if((Ply & 1)==0) { i=BKi; } else { i=WKi; }
  if(IsSquareAttackedBy(i,Ply & 1)==-1) { return(-1); }

    
/***** MACROS **********************************/
#define NextDepth(d) \
  switch(RecursiveSearch(Ply+1)) { \
      case -1: break; \
      case 1: InsertMove(i,d,0); NbLegalMoves++; NbMates[Ply]++; break; \
      case 2: InsertMove(i,d,1); NbLegalMoves++; NbStalemates[Ply]++; break; \
      case 0: /* InsertMove(i,d,2);*/ NbLegalMoves++; break; \
  }

/* macro used to generate moves for K,N,B,R,Q */
/* see also P and castling */
#define GenerateMove(Displacement,OppositeColorTest,t) \
  t=Board[i+Displacement]; \
  if((t==0) || ((t & (1+2))==OppositeColorTest)) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+Displacement+120]=Board[i]; \
    Board[i+120]=ES; \
    NextDepth(Displacement) \
  }
/***** END MACROS ******************************/
    
  /* if we are not at MaxPly, generate all possible moves */
  if(Ply<MaxPly) {
    OppositePlayer=1-(Ply&1)+2;
    NbLegalMoves=0;
    i=120*Ply+21;
    for(y=0;y<8;y++) {
      for(x=0;x<8;x++) {
        /* Board[i] contains a piece of current's player color? */
        if((Board[i]&(2+1))==(2+(Ply&1))) {
          switch(Board[i]&254) {
            /**** generate moves for P ****/
            /*!!! PROMOTION: blanc et dans [31;38] ou noir et dans [81;88] */
            /*!!! il faut le tester dans single step + capture */
            case WP:
              d=2*(Ply&1)-1; /* direction of pawn */
              temp=i%120; /* position in board */
              /* single step (and maybe double step) */
              if(Board[i+10*d]==ES) {
                CopyPlyAToPlyB(Ply,Ply+1);
                Board[i+10*d+120]=Board[i];
                Board[i+120]=ES;
                NextDepth(10*d)
                /* double step? */
                if(  ((temp>=81) && (d==-1) && (Board[i+20*d]==ES))
                   ||((temp<=38) && (d== 1) && (Board[i+20*d]==ES)) ) {
                  CopyPlyAToPlyB(Ply,Ply+1);
                  Board[i+20*d+120]=Board[i];
                  Board[i+120]=ES;
                  ep[Ply+1]=temp+10*d; /* set ep right */
                  NextDepth(20*d)
                  ep[Ply+1]=0; /* restore usual ep right */
                }
              }
              /* captures by pawn */
              if((Board[i+9*d] & (2+1))==2+1-(Ply&1)) {
                CopyPlyAToPlyB(Ply,Ply+1);
                Board[i+9*d+120]=Board[i];
                Board[i+120]=ES;
                NextDepth(9*d)
              }
              if((Board[i+11*d] & (2+1))==2+1-(Ply&1)) {
                CopyPlyAToPlyB(Ply,Ply+1);
                Board[i+11*d+120]=Board[i];
                Board[i+120]=ES;
                NextDepth(11*d)
              }
              /* en passant captures */
              if(temp+9*d==ep[Ply]) {
                CopyPlyAToPlyB(Ply,Ply+1);
                Board[i+9*d+120]=Board[i];
                Board[i+(9-10)*d+120]=ES;
                Board[i+120]=ES;
                switch(RecursiveSearch(Ply+1)) {
                    case -1: break;
                    case 1: InsertMove(i,9*d,2); NbLegalMoves++; NbMates[Ply]++; break;
                    case 2: InsertMove(i,9*d,3); NbLegalMoves++; NbStalemates[Ply]++; break;
                    case 0: NbLegalMoves++; break;
                }
              }
              if(temp+11*d==ep[Ply]) {
                CopyPlyAToPlyB(Ply,Ply+1);
                Board[i+11*d+120]=Board[i];
                Board[i+(11-10)*d+120]=ES;
                Board[i+120]=ES;
                switch(RecursiveSearch(Ply+1)) {
                    case -1: break;
                    case 1: InsertMove(i,11*d,2); NbLegalMoves++; NbMates[Ply]++; break;
                    case 2: InsertMove(i,11*d,3); NbLegalMoves++; NbStalemates[Ply]++; break;
                    case 0: NbLegalMoves++; break;
                }
              }
              break;
            /**** generate moves for K ****/
            /*!!! actualise castling rights */
            case WK:
              GenerateMove(-1,OppositePlayer,temp)
              GenerateMove(-11,OppositePlayer,temp)
              GenerateMove(-10,OppositePlayer,temp)
              GenerateMove(-9,OppositePlayer,temp)
              GenerateMove(+1,OppositePlayer,temp)
              GenerateMove(+11,OppositePlayer,temp)
              GenerateMove(+10,OppositePlayer,temp)
              GenerateMove(+9,OppositePlayer,temp)
              break;
            /**** generate moves for N ****/
            case WN:
              GenerateMove(-8,OppositePlayer,temp)
              GenerateMove(-12,OppositePlayer,temp)
              GenerateMove(-19,OppositePlayer,temp)
              GenerateMove(-21,OppositePlayer,temp)
              GenerateMove(+21,OppositePlayer,temp)
              GenerateMove(+19,OppositePlayer,temp)
              GenerateMove(+12,OppositePlayer,temp)
              GenerateMove(+8,OppositePlayer,temp)
              break;
            default: 
              /**** generate moves for B/Q ****/
              if((Board[i]&WB)==WB) {
                d=0; do{ d+=9; GenerateMove(d,OppositePlayer,temp); }while(Board[i+d]==0);
                d=0; do{ d+=11; GenerateMove(d,OppositePlayer,temp); }while(Board[i+d]==0);
                d=0; do{ d-=9; GenerateMove(d,OppositePlayer,temp); }while(Board[i+d]==0);
                d=0; do{ d-=11; GenerateMove(d,OppositePlayer,temp); }while(Board[i+d]==0);
              }
              /**** generate moves for R/Q ****/
              /*!!! actualise castling rights if not Q */
              if((Board[i]&WR)==WR) {
                d=0; do{ d+=1; GenerateMove(d,OppositePlayer,temp); }while(Board[i+d]==0);
                d=0; do{ d+=10; GenerateMove(d,OppositePlayer,temp); }while(Board[i+d]==0);
                d=0; do{ d-=1; GenerateMove(d,OppositePlayer,temp); }while(Board[i+d]==0);
                d=0; do{ d-=10; GenerateMove(d,OppositePlayer,temp); }while(Board[i+d]==0);
              }
            break;
          }
        }
        i++;
      } /* next x */
      i+=2;
    } /* next y*/
    /*!!! castlings, will use IsSquareAttackedBy(i,c) */
    /*!!! printf("nb of legal moves:%d\n",NbLegalMoves);  */

    /* check if previous move mated, stalemated or nothing */  
    if(NbLegalMoves==0) {
      /* make sure that K of current player is not in check */
      if((Ply & 1)==0) { i=WKi; } else { i=BKi; }
      if(IsSquareAttackedBy(i,1-(Ply & 1))==-1) {
        return(1);
      } else {
        return(2);
      }
    } else {
      NbGames[Ply]+=(double) NbLegalMoves;
      return(0);
    }
  }
  return(0);
}

/**************************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  int c,i,x,y;
  setbuf(stdout, NULL); /* no buffering for output! */
  
  /* initialise arrays */
  for(c=0;c<MaxPly+1;c++) {
    ep[c]=0;      /* no ep capture possible */
    castle[c]=63; /* no K or R has moved */
    NbGames[c]=0.0;
    NbMates[c]=0;
    NbStalemates[c]=0;
  }
  ep[MaxPly+1]=0; /*!!! useless? */
  
  /* initialise Board for Ply 0 */
  for(c=0;c<(MaxPly+1)*12*10;c++) { Board[c]=FS; } /* forbidden squares*/
  i=21;for(y=0;y<8;y++){for(x=0;x<8;x++){Board[i++]=ES;}i+=2;} /* empty squares */
 
  Board[21]=BR;Board[22]=BN;Board[23]=BB;Board[24]=BQ;
  Board[25]=BK;Board[26]=BB;Board[27]=BN;Board[28]=BR;
  for(x=0;x<8;x++) { Board[31+x]=BP; }
  for(x=0;x<8;x++) { Board[81+x]=WP; }  
  Board[91]=WR;Board[92]=WN;Board[93]=WB;Board[94]=WQ;
  Board[95]=WK;Board[96]=WB;Board[97]=WN;Board[98]=WR;

/*!!!  Board[71]=WN; Board[62]=BP;*/
/*!!!  PrintPly(0); printf("\n");*/
 
  /* go */
  RecursiveSearch(0);
  
  /* Print the number of games */
  printf("ply;    NbGames   ; NbMates ; NbStalemates\n");
  for(c=0;c<MaxPly;c++) {
    if(c!=MaxPly-1) {
      printf("%3d;%14.0lf;%9d;%9d\n",c+1,NbGames[c],NbMates[c],NbStalemates[c]);
    } else {
      printf("%3d;%14.0lf; untested; untested\n",c+1,NbGames[c]);
    }  
  }
  
  /* Print the ending moves */
  printf("\nEnding moves:\n");
  for(c=0;c<NbEndMoves;c++) {
    x=EndMoves[c*2];
    y=(x>>26) & 63; /* y=ply */
    printf("%d.",(y/2)+1); if((y&1)==1) { printf(".. "); }
    y=(x>>20) & 62; /* y=piece */
    switch(y) {
      case WK: printf("K"); break;
      case WP: printf("P"); break;
      case WN: printf("N"); break;
      case WB: printf("B"); break;
      case WR: printf("R"); break;
      case WQ: printf("Q"); break;
      default: printf("***error1***");
    }
    y=((x>>14)&127) - 21; /* y=square */
    printf("%c%d",y%10+'a',8-y/10);
    y=(x>>8) & 62; /* y=piece */
    if((y==0) && ((x&3)<2)) {
      printf("-");
    } else {
      printf("x");
      switch(y) {
        case WK: printf("K"); break;
        case ES:case WP: printf("P"); break; /* also en passant capture */
        case WN: printf("N"); break;
        case WB: printf("B"); break;
        case WR: printf("R"); break;
        case WQ: printf("Q"); break;
        default: printf("***error2***");
      }
    }
    y=((x>>2)&127) - 21; /* y=square */
    printf("%c%d",y%10+'a',8-y/10);
    switch(x&3) {
      case 0: printf("#"); break;
      case 1: printf("="); break;
      case 2: printf("ep#"); break;
      case 3: printf("ep="); break;
    }
    printf(" ; %d\n",EndMoves[c*2+1]);
  }
  
  exit(0);
}
