/* Search for (short) uniquely realisable mate in chess variants */
/* Alain Brobecker - 2008 august-september */
/* 2008.09.21 - Added promotion. 
                Conditionnal compilation of castling.
                Added a mask in PrintMove function. */
/* 2008.09.20 - Added castling. */
/* 2008.09.13 - Added recognition of minimal notation for a problem.
                Counting nb of EndMoves for every ply.
                Sorted EndMoves according to Ply and Occurences. */
/* 2008.09.02 - Removed BUG for test of mate at MaxPly when K is in check,
                it only was in initialisation of Boards with FS!!! */
/* 2008.08.28 - Set variables in CopyPlyAToPlyB as global.
                Tested if opposite K attacks square in IsSquareAttacked (for castling).
                Cleaned up the code.  */
/* 2008.08.26 - Inserted a dir variable so that it's not computed everytime. */
/* 2008.08.17 - Set WKx,WKy,BKx,BKy as global variables.
                Added test of mate at MaxPly when K is in check, but it BUGS?! */
/* 2008.08.15 - Added (simple) recognition of similar EndMoves. */
/* 2008.08.14 - Added move generation for Pawns (ep but no promotion). */
/* 2008.08.13 - Added function for printing moves.
                Added mate/stalemate recognition.
                Added move generation for Kings, Rooks, Bishops, Queens. */
/* 2008.08.12 - Added move generation for kNights, with a macro. */
/* 2008.05.23 - Modified IsKingInCheck(i) to IsSquareAttackedBy(i,c) so that
                it can be used to check castling. */
/* 2008.05.18 - Added IsKingInCheck(i). */
/* 2008.05.17 - Start+speed tests: dim1 arrays are faster than dim2. */
#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

/* #define CASTLING CASTLING */ /* remove if you don't want castling to be implemented */

#define MaxPly 5
#define MaxNbEndMoves 9999

#define FALSE 0
#define TRUE 1

/* c is for color (0=White, 1=Black), must be lowest bit. */
/* 2nd bit is set for pieces. And Q=B+R. */
#define ES 0  /* 000000 empty square */
#define FS 1  /* 000001 forbidden squares around the board */
#define WK 2  /* 00001c K */
#define BK 3
#define WP 6  /* 00011c P */
#define BP 7
#define WN 10 /* 00101c N */
#define BN 11
#define WB 18 /* 01001c B */
#define BB 19
#define WR 34 /* 10001c R */
#define BR 35
#define WQ 50 /* 11001c Q */
#define BQ 51

/* Every Board is made of 21+8+2+8+2+8+2+8+2+8+2+8+2+8+2+8+21=120 chars */
/* The 21s and 2s are used for forbidden squares (FS) around the board */
/* We could gain space by having 10+MaxPly*(11+8+2+...+2+8+11)+10 instead */

unsigned char Board[(MaxPly+2)*12*10];
unsigned int epSquare[MaxPly+3]; /* square where an ep capture is possible or 0 */
unsigned int NbMates[MaxPly+2],NbStalemates[MaxPly+2];
unsigned int EndMove[MaxNbEndMoves],EndMoveOccurence[MaxNbEndMoves];
unsigned int NbEndMoves=0;
unsigned int NbEndMovesAtPly[MaxPly+2];
double NbGames[MaxPly+2];
#ifdef CASTLING
  unsigned char CastlingRights[MaxPly+3]; /*6 bits:BQR,WQR,BKR,WKR,BK,WK, 1 when it has moved*/
#endif

/* Print 8*8 central part of Board[Ply], mostly for debugging */
void PrintPly(Ply) int Ply; {
  int i,x,y;
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i++]) {
        case ES: printf("."); break; case FS: printf("*"); break;
        case WK: printf("K"); break; case BK: printf("k"); break;
        case WP: printf("P"); break; case BP: printf("p"); break;
        case WN: printf("N"); break; case BN: printf("n"); break;
        case WB: printf("B"); break; case BB: printf("b"); break;
        case WR: printf("R"); break; case BR: printf("r"); break;
        case WQ: printf("Q"); break; case BQ: printf("q"); break;
        default:printf("?"); break; /* error */
      }
    }
    printf("\n");
    i+=2;
  }
}

/* same as above, but print it in FEN notation */
void PrintFEN(Ply) int Ply; {
  int i,x,y,nb;
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    nb=0;
    for(x=0;x<8;x++) {
      if(Board[i]==ES) {
        nb++;
      } else {
        if(nb!=0) { printf("%d",nb); nb=0; }
        switch(Board[i]) {
          case FS: printf("*"); break;
          case WK: printf("K"); break; case BK: printf("k"); break;
          case WP: printf("P"); break; case BP: printf("p"); break;
          case WN: printf("N"); break; case BN: printf("n"); break;
          case WB: printf("B"); break; case BB: printf("b"); break;
          case WR: printf("R"); break; case BR: printf("r"); break;
          case WQ: printf("Q"); break; case BQ: printf("q"); break;
          default:printf("?"); break; /* error */
        }
      }
      i++;
    }
    if(nb!=0) { printf("%d",nb); }
    if(y!=7) { printf("/"); }
    i+=2;
  }
}

/* used to print the ending move in standard format */
void PrintMove(Move) unsigned int Move; {
  int t;
  t=(Move>>26) & 63;
/*!!!    for(i=0;i<y;i++) { printf("  "); }*/
  printf("%d.",(t/2)+1); if((t&1)==1) { printf(".. "); }
  t=(Move>>20) & 62; /* piece without color */
  switch(t) {
    case WK: printf("K"); break;
    case WP: printf("P"); break;
    case WN: printf("N"); break;
    case WB: printf("B"); break;
    case WR: printf("R"); break;
    case WQ: printf("Q"); break;
    default: printf("***error1***");
  }
  t=((Move>>14)&127) - 21; /* square */
  printf("%c%d",t%10+'a',8-t/10);
  t=(Move>>8) & 62; /* piece without color */
  if((t==0) && ((Move&3)<2)) {
    printf("-");
  } else {
    printf("x");
    switch(t) {
      case WK: printf("K"); break;
      case ES:case WP: printf("P"); break; /* also en passant capture */
      case WN: printf("N"); break;
      case WB: printf("B"); break;
      case WR: printf("R"); break;
      case WQ: printf("Q"); break;
      default: printf("***error2***");
    }
  }
  t=((Move>>2)&127) - 21; /* square */
  printf("%c%d",t%10+'a',8-t/10);
  switch(Move&3) {
    case 0: printf("#"); break;
    case 1: printf("="); break;
    case 2: printf("*#"); break; /* special move: promotion, ep, castling */
    case 3: printf("*="); break;
  }
}

/* Insert move from square index to square index+displacement into list */
unsigned int IM_Ply,IM_Move,IM_t,IM_Found;
void InsertMove(Index,Displacement,Type) int Index,Displacement,Type; {
  IM_Ply=Index/120;
  IM_Move=IM_Ply<<26;                   /* ply (6 bits) */
  IM_Move+=(Board[Index]&254)<<(21-1);  /* piece (5 bits) */
  IM_Move+=(Index-120*IM_Ply)<<14;      /* square in [0;119] (7 bits) */
  IM_Move+=(Board[Index+Displacement]&254)<<(9-1); /* piece (5 bits) */
  IM_Move+=(Index+Displacement-120*IM_Ply)<<2; /* square (7 bits) */
  IM_Move+=Type; /* 0=mate, 1=stalemate, 2=ep mate, 3=ep stalemate (2 bits) */
  IM_Found=FALSE;
  IM_t=0;
  while((IM_t<NbEndMoves) && (IM_Found==FALSE)) {
    if(IM_Move==EndMove[IM_t]) {
      IM_Found=TRUE;
      EndMoveOccurence[IM_t]++;
    }
    IM_t++;
  }
  if(IM_Found==FALSE) {
    PrintMove(IM_Move); printf(";"); PrintFEN(IM_Ply); printf("\n");
    EndMove[NbEndMoves]=IM_Move;
    EndMoveOccurence[NbEndMoves]=1;
    NbEndMoves++;
    if(NbEndMoves>=MaxNbEndMoves) {
      printf("***error: More than %d EndMove!\n",MaxNbEndMoves); exit(-1);
    }
  }
}


/* Copy Board[PlyA] to Board[PlyB] */
int CP_i,CP_j,CP_x,CP_y;
void CopyPlyAToPlyB(PlyA,PlyB) int PlyA,PlyB; {
  CP_j=120*PlyB+21; CP_i=120*PlyA+21;
  for(CP_y=0;CP_y<8;CP_y++) {
    for(CP_x=0;CP_x<8;CP_x++) { Board[CP_j++]=Board[CP_i++]; }
    CP_j+=2; CP_i+=2;
  }
#ifdef CASTLING
  CastlingRights[PlyB]=CastlingRights[PlyA]; /* copy castling rights */
#endif
}

/**************************************************************************/
/* Returns TRUE if square "index" is attacked by a "color" piece, else    */
/* it returns FALSE */

#define IsSquareAttackeByKingOrRider(dir,piece) \
  ISA_i=index+dir; \
  if(Board[ISA_i]==(WK | color)) { \
    return(TRUE); \
  } else { \
    while(Board[ISA_i]==ES) { ISA_i+=dir; } \
    if((Board[ISA_i] & (piece+1))==(piece | color)) { return(TRUE); } \
  }

int ISA_i;
int IsSquareAttackedBy(index,color) int index,color; {
  /* attacked by opposite color N? */
  if(Board[index-8]==(WN | color)) { return(TRUE); }
  if(Board[index-12]==(WN | color)) { return(TRUE); }
  if(Board[index-19]==(WN | color)) { return(TRUE); }
  if(Board[index-21]==(WN | color)) { return(TRUE); }
  if(Board[index+21]==(WN | color)) { return(TRUE); }
  if(Board[index+19]==(WN | color)) { return(TRUE); }
  if(Board[index+12]==(WN | color)) { return(TRUE); }
  if(Board[index+8]==(WN | color)) { return(TRUE); }
  /* attacked by opposite color P? */
  if(Board[index+9-18*color]==(WP | color)) { return(TRUE); }
  if(Board[index+11-22*color]==(WP | color)) { return(TRUE); }
  /* attacked by opposite color K, B or Q? */
  IsSquareAttackeByKingOrRider(9,WB)
  IsSquareAttackeByKingOrRider(11,WB)
  IsSquareAttackeByKingOrRider(-11,WB)
  IsSquareAttackeByKingOrRider(-9,WB)
  /* attacked by opposite color R or Q? */
  IsSquareAttackeByKingOrRider(1,WR)
  IsSquareAttackeByKingOrRider(10,WR)
  IsSquareAttackeByKingOrRider(-10,WR)
  IsSquareAttackeByKingOrRider(-1,WR)
  return(FALSE);
}

/**************************************************************************/
/* Returns:
 * -1 if K of previous player is still in check (previous move was illegal)
 *  1 if K of current player is mate
 *  2 if K of current player is stalemate 
 *  0 otherwise. */
   
#define NextDepth(d,type) \
  switch(RecursiveSearch(Ply+1)) { \
      case 0: /*!!! InsertMove(i,d,1);*/ NbLegalMoves++; break; \
      case -1: break; \
      case 1: InsertMove(i,d,type); NbLegalMoves++; NbMates[Ply]++; break; \
      case 2: InsertMove(i,d,type+1); NbLegalMoves++; NbStalemates[Ply]++; break; \
  }

/* macro used to generate moves for Royal Leaper */
#define GenerateMoveK(Displacement,t) \
  t=Board[i+Displacement]; \
  if((t==ES) || ((t & (1+2))==OppositePlayer)) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+Displacement+120]=Board[i]; \
    Board[i+120]=ES; \
    NextDepth(Displacement,0) \
  }

/* macro used to generate moves for Leaper */
#define GenerateMoveLeaper(Displacement,t) \
  t=Board[i+Displacement]; \
  if((t==ES) || ((t & (1+2))==OppositePlayer)) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+Displacement+120]=Board[i]; \
    Board[i+120]=ES; \
    NextDepth(Displacement,0) \
  }
  
/* macro used to generate moves for Rider */
#define GenerateRiderMove(dir,t) \
  t=dir; \
  while(Board[i+t]==ES) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+t+120]=Board[i]; \
    Board[i+120]=ES; \
    NextDepth(t,0) \
    t+=dir; \
  } \
  if((Board[i+t] & (1+2))==OppositePlayer) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+t+120]=Board[i]; \
    Board[i+120]=ES; \
    NextDepth(t,0) \
  }  
   
int WKx,WKy,BKx,BKy;
int RecursiveSearch(Ply) int Ply; {
  int i,x,y,OppositePlayer,CurrentPlayer,NbLegalMoves,dir,t0,t1;
#ifdef CASTLING  
  int OldCastlingRights;
#endif  
  
  /* search Ks' positions and check distance between both */
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i]) {
        case WK: WKx=x; WKy=y; t0=i; break; /* t0=index fo WK */
        case BK: BKx=x; BKy=y; t1=i; break; /* t1=index fo BK */
      }
      i++;
    }
    i+=2;
  }
  x=WKx-BKx; if(x<0) { x=-x; }
  y=WKy-BKy; if(y<0) { y=-y; }
  if((x<2) && (y<2)) { return(-1); /*!!! printf("dist WK-BK: (%d;%d)\n",x,y);*/ }
  
  /* Set t0=index of current player's K, t1=index of previous player's K,
     then see if previous player's K is still in check (move was illegal). */
  CurrentPlayer=Ply&1;
  if(CurrentPlayer==1) { t1=t1-t0; t0=t0+t1; t1=t0-t1; }
  if(IsSquareAttackedBy(t1,CurrentPlayer)==TRUE) { return(-1); }
  
  /* Set t0 to TRUE if current player's K is in check. Then quit recursion 
     according to depth. At MaxPly we only process cases were K is in check. */
  if(Ply>MaxPly) { return(0); }
  OppositePlayer=1-CurrentPlayer;
  t0=IsSquareAttackedBy(t0,OppositePlayer);
  if((Ply==MaxPly) && (t0==FALSE)) { return(0); }

  /* We must process this Ply, so generate all possible moves. */
  dir=2*CurrentPlayer-1; /* direction of pawns */
  CurrentPlayer+=2;
  OppositePlayer+=2;
  NbLegalMoves=0;
  i=i-80; /* 120*Ply+21; */
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      /* Board[i] contains a piece of current's player color? */
      if((Board[i]&(2+1))==CurrentPlayer) {
        switch(Board[i]&254) {
          /**** generate moves for P ****/
          /*!!! PROMOTION: (blanc et y=1) ou (noir et y=6) */
          /*!!! il faut le tester dans single step + capture */
          case WP:
            /* single step (and maybe double step) */
            if(Board[i+10*dir]==ES) {
              CopyPlyAToPlyB(Ply,Ply+1);
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+120]=ES;
                Board[i+10*dir+120]=WN+CurrentPlayer-2; /* to N */
                NextDepth(10*dir,2)
                Board[i+10*dir+120]=WB+CurrentPlayer-2; /* to B */
                NextDepth(10*dir,2)
                Board[i+10*dir+120]=WR+CurrentPlayer-2; /* to R */
                NextDepth(10*dir,2)
                Board[i+10*dir+120]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth(10*dir,2)
              } else {
                /* no promotion */
                Board[i+10*dir+120]=Board[i];
                Board[i+120]=ES;
                NextDepth(10*dir,0)
                /* double step? */
                if(  ((y==6) && (dir==-1) && (Board[i+20*dir]==ES))
                   ||((y==1) && (dir== 1) && (Board[i+20*dir]==ES)) ) {
                  CopyPlyAToPlyB(Ply,Ply+1);
                  Board[i+20*dir+120]=Board[i];
                  Board[i+120]=ES;
                  epSquare[Ply+1]=i+10*dir+120; /* set ep right */
                  NextDepth(20*dir,0)
                  epSquare[Ply+1]=0; /* restore usual ep right */
                }
              }
            }
            /* captures by pawn */
            if((Board[i+9*dir] & (2+1))==OppositePlayer) {
              CopyPlyAToPlyB(Ply,Ply+1);
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+120]=ES;
                Board[i+9*dir+120]=WN+CurrentPlayer-2; /* to N */
                NextDepth(9*dir,2)
                Board[i+9*dir+120]=WB+CurrentPlayer-2; /* to B */
                NextDepth(9*dir,2)
                Board[i+9*dir+120]=WR+CurrentPlayer-2; /* to R */
                NextDepth(9*dir,2)
                Board[i+9*dir+120]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth(9*dir,2)
              } else {
                /* no promotion */
                Board[i+9*dir+120]=Board[i];
                Board[i+120]=ES;
                NextDepth(9*dir,0)
              }
            }
            if((Board[i+11*dir] & (2+1))==OppositePlayer) {
              CopyPlyAToPlyB(Ply,Ply+1);
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+120]=ES;
                Board[i+11*dir+120]=WN+CurrentPlayer-2; /* to N */
                NextDepth(11*dir,2)
                Board[i+11*dir+120]=WB+CurrentPlayer-2; /* to B */
                NextDepth(11*dir,2)
                Board[i+11*dir+120]=WR+CurrentPlayer-2; /* to R */
                NextDepth(11*dir,2)
                Board[i+11*dir+120]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth(11*dir,2)
              } else {
                /* no promotion */
                Board[i+11*dir+120]=Board[i];
                Board[i+120]=ES;
                NextDepth(11*dir,0)
              }
            }
            /* en passant captures */
            if(i+9*dir==epSquare[Ply]) {
              CopyPlyAToPlyB(Ply,Ply+1);
              Board[i+9*dir+120]=Board[i];
              Board[i+(9-10)*dir+120]=ES;
              Board[i+120]=ES;
              NextDepth(9*dir,2)
            }
            if(i+11*dir==epSquare[Ply]) {
              CopyPlyAToPlyB(Ply,Ply+1);
              Board[i+11*dir+120]=Board[i];
              Board[i+(11-10)*dir+120]=ES;
              Board[i+120]=ES;
              NextDepth(11*dir,2)
            }
          break;
          /**** generate moves for K ****/
          /*!!! actualise castling rights */
          case WK:
#ifdef CASTLING
            OldCastlingRights=CastlingRights[Ply];
            CastlingRights[Ply]=OldCastlingRights | (CurrentPlayer-2+1); /* set bit for BK or WK to 1 */
#endif
            GenerateMoveK(-1,t1)
            GenerateMoveK(-11,t1)
            GenerateMoveK(-10,t1)
            GenerateMoveK(-9,t1)
            GenerateMoveK(+1,t1)
            GenerateMoveK(+11,t1)
            GenerateMoveK(+10,t1)
            GenerateMoveK(+9,t1)
#ifdef CASTLING
            CastlingRights[Ply]=OldCastlingRights;
#endif
          break;
          /**** generate moves for N ****/
          case WN:
            GenerateMoveLeaper(-8,t1)
            GenerateMoveLeaper(-12,t1)
            GenerateMoveLeaper(-19,t1)
            GenerateMoveLeaper(-21,t1)
            GenerateMoveLeaper(+21,t1)
            GenerateMoveLeaper(+19,t1)
            GenerateMoveLeaper(+12,t1)
            GenerateMoveLeaper(+8,t1)
          break;
          default:
            /**** generate moves for N/Q ****/
            if((Board[i]&WB)==WB) {
              GenerateRiderMove(9,t1)
              GenerateRiderMove(11,t1)
              GenerateRiderMove(-11,t1)
              GenerateRiderMove(-9,t1)
            }
            /**** generate moves for R/Q ****/
            if((Board[i]&WR)==WR) {
#ifdef CASTLING
              /* actualise castling rights if R */
              if((Board[i]&254)==WR) {
                OldCastlingRights=CastlingRights[Ply];
                if((i%120)==91-70*(CurrentPlayer-2)) {
                  CastlingRights[Ply]=OldCastlingRights | ((CurrentPlayer-2+1)*16);
                }
                if((i%120)==98-70*(CurrentPlayer-2)) {
                  CastlingRights[Ply]=OldCastlingRights | ((CurrentPlayer-2+1)*4);
                }
              }
#endif
              GenerateRiderMove(1,t1)
              GenerateRiderMove(10,t1)
              GenerateRiderMove(-10,t1)
              GenerateRiderMove(-1,t1)
#ifdef CASTLING
              if((Board[i]&254)==WR) {
                CastlingRights[Ply]=OldCastlingRights;
              }
#endif
            }
          break;
        }
      }
      i++;
    } /* next x */
    i+=2;
  } /* next y */
  
  /**** generate castlings if possible ****/
  /* It is possible to make a simpler handling of castling if Ply<=8 by */
  /* just checking if K and R are on their normal square, but this will */
  /* fail if ply>8, eg 1.Nh3 b6 2.Rg1 Ba6 3.Rh1 Bxg2 4.Bxg2 ~ 5.oo      */
  
#ifdef CASTLING
  if((CastlingRights[Ply] & (5<<(CurrentPlayer-2)))==0) { /* kingside castling */
    i=120*Ply+95-70*(CurrentPlayer-2); /* points on K of CurrentPlayer */
    if((Board[i+1]==ES) && (Board[i+2]==ES)
       && (IsSquareAttackedBy(i,OppositePlayer-2)==FALSE)
       && (IsSquareAttackedBy(i+1,OppositePlayer-2)==FALSE)
       && (IsSquareAttackedBy(i+2,OppositePlayer-2)==FALSE)) {
      CopyPlyAToPlyB(Ply,Ply+1);
      Board[i+2+120]=Board[i]; Board[i+120]=ES; /* move K */
      Board[i+1+120]=Board[i+3]; Board[i+120+3]=ES; /* move R */
      NextDepth(2,2)
    }
  }
  if((CastlingRights[Ply] & (17<<(CurrentPlayer-2)))==0) { /* queenside castling */
    i=120*Ply+95-70*(CurrentPlayer-2);
    if((Board[i-1]==ES) && (Board[i-2]==ES) && (Board[i-3]==ES)
       && (IsSquareAttackedBy(i,OppositePlayer-2)==FALSE)
       && (IsSquareAttackedBy(i-1,OppositePlayer-2)==FALSE)
       && (IsSquareAttackedBy(i-2,OppositePlayer-2)==FALSE)) {
      CopyPlyAToPlyB(Ply,Ply+1);
      Board[i-2+120]=Board[i]; Board[i+120]=ES; /* move K */
      Board[i-1+120]=Board[i-4]; Board[i+120-4]=ES; /* move R */
      NextDepth(-2,2)
    }
  }
#endif

  /* check if previous move mated, stalemated or nothing */  
  if(NbLegalMoves==0) {
    if(t0==TRUE) { return(1); } else { return(2); }
  } else {
    NbGames[Ply]+=(double) NbLegalMoves;
    return(0);
  }
}

/**************************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  int c,i,x,y;
  setbuf(stdout, NULL); /* no buffering for output! */
  
  /* initialise arrays */
#ifdef CASTLING
  CastlingRights[0]=0; /* copied from previous ply */
#endif  
  for(c=0;c<MaxPly+2;c++) {
    epSquare[c]=0;      /* no ep capture possible */
    NbGames[c]=0.0;
    NbMates[c]=0;
    NbStalemates[c]=0;
    NbEndMovesAtPly[c]=0;
  }
  epSquare[MaxPly+2]=0; /*!!! useless? */
  
  /* initialise Board for Ply 0 */
  for(c=0;c<(MaxPly+2)*12*10;c++) { Board[c]=FS; } /* forbidden squares*/
  i=21;for(y=0;y<8;y++){for(x=0;x<8;x++){Board[i++]=ES;}i+=2;} /* empty squares */
  Board[21]=BR;Board[22]=BN;Board[23]=BB;Board[24]=BQ;
  Board[25]=BK;Board[26]=BB;Board[27]=BN;Board[28]=BR;
  for(x=0;x<8;x++) { Board[31+x]=BP; }
  for(x=0;x<8;x++) { Board[81+x]=WP; }  
  Board[91]=WR;Board[92]=WN;Board[93]=WB;Board[94]=WQ;
  Board[95]=WK;Board[96]=WB;Board[97]=WN;Board[98]=WR;

/*  Board[48]=WP; Board[88]=ES; *//* 1.h4 d5 2.h5 Nd7 3.h6 Ndf6 ... */
/*  Board[54]=BP; Board[34]=ES; *//* to test mate by 6.gxf8=N# */
/*  Board[46]=BN; Board[22]=ES; *//* Peter Rsler, Problemkiste 1994/08 */
/*  PrintPly(0); printf("\n"); */
  
  /* go */
  RecursiveSearch(0);
  
  /* sort (stupidly) the EndMove according to their occurence */
  for(x=0;x<NbEndMoves-1;x++) {
    i=x;
    c=EndMoveOccurence[x];
    for(y=x+1;y<NbEndMoves;y++) {
      if(EndMoveOccurence[y]<c) { c=EndMoveOccurence[y]; i=y; }
    }
    c=EndMove[x];
    EndMove[x]=EndMove[i];
    EndMove[i]=c;
    c=EndMoveOccurence[x];
    EndMoveOccurence[x]=EndMoveOccurence[i];
    EndMoveOccurence[i]=c;
  }
  
  /* Print the ending moves */
  printf("\nEnding moves;occurences\n");
  for(i=0;i<MaxPly+1;i++) {
    for(c=0;c<NbEndMoves;c++) {
      x=EndMove[c];
      y=(x>>26);
      if(y==i) {
        NbEndMovesAtPly[y]++;
        PrintMove(x);
        printf(";%d\n",EndMoveOccurence[c]);
      }
    }
  }
  
  /* Print the number of games */
  printf("\nply;    NbGames   ;NbEndMoves; NbMates ; NbStalemates\n");
  for(c=0;c<MaxPly;c++) {
    if(c!=MaxPly-1) {
      printf("%3d;%14.0lf;%10d;%9d;%9d\n",c+1,NbGames[c],NbEndMovesAtPly[c],NbMates[c],NbStalemates[c]);
    } else {
      printf("%3d;%14.0lf;%10d;%9d; untested\n",c+1,NbGames[c],NbEndMovesAtPly[c],NbMates[c]);
    }  
  }
  
  /* search for minimal information for every unique EndMove */
  printf("\nPossible problems are:\n");
  unsigned int NbMinimal[8]; /* 8 types of minimal information, see below */
  c=0;
  while((EndMoveOccurence[c]==1) && (c<NbEndMoves)) {
    x=EndMove[c];
    for(i=0;i<8;i++) { NbMinimal[i]=0; }
    for(i=0;i<NbEndMoves;i++) {
      y=EndMove[i];
      /* Same ply, same playing piece, same type of endmove? */
      if( (x&((2047<<21)+3))==(y&((2047<<21)+3)) ) {
        NbMinimal[0]++; /* n.piece# */
        /* Same destination square? */
        if( (x&(127<<2))==(y&(127<<2)) ) {
          NbMinimal[4]++; /* n.piece?square# */
        }
        /* Both are moves without capture? */
        if( ((x&(31<<9))==0) && ((y&(31<<9))==0) ) {
          NbMinimal[1]++; /* n.piece-# */
          /* Same destination square? */
          if( (x&(127<<2))==(y&(127<<2)) ) {
            NbMinimal[5]++; /* n.piece-square# */
          }
        } else {
          /* Both are captures? */
          if( ((x&(31<<9))!=0) && ((y&(31<<9))!=0) ) {
            NbMinimal[2]++; /* n.piecex# */
            /* Same piece captured? */
            if( (x&(31<<9))==(y&(31<<9)) ) {
              NbMinimal[3]++; /* n.piecexpiece# */
              /* Also same destination square?*/
              if( (x&(127<<2))==(y&(127<<2)) ) {
                NbMinimal[7]++; /* n.piecexpiece_square# */
              }
            }
            /* Same destination square?*/
            if( (x&(127<<2))==(y&(127<<2)) ) {
              NbMinimal[6]++; /* n.piecexsquare# */
            }
          }
        }
      }
    }
    /* Print minimal info for this move */
    PrintMove(x); printf("; ");
    i=0;
    while((NbMinimal[i]!=1) && (i<8)) { i++; }
    switch(i) {
      case 0:printf("n.piece#"); break;
      case 1:printf("n.piece-#"); break;
      case 2:printf("n.piecex#"); break;
      case 3:printf("n.piecexpiece#"); break;
      case 4:printf("n.piece?square#"); break;
      case 5:printf("n.piece-square#"); break;
      case 6:printf("n.piecexsquare#"); break;
      case 7:printf("n.piecexpiece_square#"); break;
      default: break;
    }
    printf("\n");
    c++;
  }
  exit(0);
}
